// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.


#include <eopcc/cloud.hpp>
#include <eopcc/metrics.hpp>
#include <eopcc/registration.hpp>
#include <iostream>

namespace eopcc {



// Registers by searching for lowest RMS error score.
Point RegisterFromRmsGrid(const RegSearchSpace& regspace, const PointCloud& input, const eopcc::GridModel<double>& truth_grid) {

  Point best;
  double min_error = std::numeric_limits<double>::max();
  double mean_error;

  {
    GridModel<double> max_grid(truth_grid.extent(), truth_grid.spacing());

    for (double x_shift = regspace.start; x_shift <= regspace.stop; x_shift += regspace.delta) {
      for (double y_shift = regspace.start; y_shift <= regspace.stop; y_shift += regspace.delta) {
        Point shift = regspace.center + Point(x_shift, y_shift, 0);
        MaxHeightGrid(max_grid, input, shift);
        double zmean = ZMeanError(max_grid, truth_grid);

        shift.z -= zmean;

        double error = eopcc::ZRMSError(max_grid, truth_grid, -zmean);

        //std::cout << "mean: " << zmean << " error: " << error << std::endl;
        if (error < min_error) {
          mean_error = zmean;
          min_error = error;
          best = shift;
        }
      }
    }
  }

  double factor = regspace.refine_factor;

  std::cout << "registration: " << min_error << ", " << mean_error << " with: " << best << " and spacing " << regspace.delta*factor << std::endl;


  if (regspace.delta < std::max(truth_grid.spacingX(), truth_grid.spacingY())*regspace.stoping_factor) {
    return best;
  }

  return RegisterFromRmsGrid(regspace.refine(regspace.refine_factor, best), input, truth_grid);
}

// Registers by searching for lowest median error score.
Point RegisterFromMedianGrid(const RegSearchSpace& regspace, const PointCloud& input, const eopcc::GridModel<double>& truth_grid) {

  Point best;
  double min_error = std::numeric_limits<double>::max();
  double mean_error;

  {
    GridModel<double> max_grid(truth_grid.extent(), truth_grid.spacing());
    std::vector<double> buffer(truth_grid.grid.size());

    for (double x_shift = regspace.start; x_shift <= regspace.stop; x_shift += regspace.delta) {
      for (double y_shift = regspace.start; y_shift <= regspace.stop; y_shift += regspace.delta) {

        Point shift = regspace.center + Point(x_shift, y_shift, 0);

        MaxHeightGrid(max_grid, input, shift);
        double zmean = ZMedianError(max_grid, truth_grid, buffer, 0, false);
        shift.z -= zmean;

        double error = eopcc::ZMedianError(max_grid, truth_grid, buffer, -zmean);
        if (error < min_error) {
          mean_error = zmean;
          min_error = error;
          best = shift;
        }
      }
    }
  }

  double factor = regspace.refine_factor;

  std::cout << "registration: " << min_error << ", " << mean_error << " with: " << best << " and spacing " << regspace.delta*factor << std::endl;

  if (regspace.delta < std::max(truth_grid.spacingX(), truth_grid.spacingY())*regspace.stoping_factor) {
    return best;
  }

  return RegisterFromMedianGrid(regspace.refine(regspace.refine_factor, best), input, truth_grid);
}

} // namespace eopcc
